package org.codehaus.xfire.message;

import java.io.OutputStream;
import java.util.Calendar;
import java.util.Date;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

import org.codehaus.xfire.XFireRuntimeException;
import org.codehaus.xfire.util.DateUtils;

/**
 * LiteralWriter
 * 
 * @author <a href="mailto:dan@envoisolutions.com">Dan Diephouse</a>
 */
public class LiteralWriter
    implements MessageWriter
{
    private XMLStreamWriter writer;
    
    private String namespace;
    
    private String name;

    private String prefix;

    private static Calendar calendar = Calendar.getInstance();

    /**
     * Create a LiteralWriter but without writing an element name.
     * 
     * @param writer
     */
    public LiteralWriter(XMLStreamWriter writer)
    {
        this.writer = writer;
    }
    
    public LiteralWriter( XMLStreamWriter writer, String name, String namespace )
    {
        this(writer, name, namespace, null);
    }

    public LiteralWriter( XMLStreamWriter writer, String name, String namespace, String prefix )
    {
        this.writer = writer;
        this.namespace = namespace;
        this.name = name;
        this.prefix = prefix;
        
        try
        {
            writeStartElement();
        }
        catch ( XMLStreamException e )
        {
            // TODO: I'm still not sure what should happen here.
            throw new XFireRuntimeException("Error writing document.", e);
        }
    }
    
    /**
     * @param os
     * @throws XMLStreamException 
     */
    public LiteralWriter(OutputStream os, String name, String namespace) 
        throws XMLStreamException
    {
        XMLOutputFactory ofactory = XMLOutputFactory.newInstance();
        this.writer = ofactory.createXMLStreamWriter(os);
        
        this.namespace = namespace;
        this.name = name;
        
        try
        {
            writeStartElement();
        }
        catch ( XMLStreamException e )
        {
            // TODO: I'm still not sure what should happen here.
            throw new XFireRuntimeException("Error writing document.", e);
        }
    }

    private void writeStartElement() 
        throws XMLStreamException
    {
        if (namespace != null)
        {
            if (prefix == null)
                prefix = "";
            
            writer.setPrefix(prefix, namespace);
            writer.writeStartElement(prefix, name, namespace);
            writer.writeNamespace(prefix, namespace);
        }
        else
        {
            writer.writeStartElement(name);
        }
    }

    /**
     * @see org.codehaus.xfire.message.MessageWriter#writeValue(java.lang.Object)
     */
    public void writeValue(Object value)
    {
        try
        {
            if ( value != null )
                writer.writeCharacters( value.toString() );
        }
        catch ( XMLStreamException e )
        {
            // TODO: I'm still not sure what should happen here.
            throw new XFireRuntimeException("Error writing document.", e);
        }
    }

    /**
     * @see org.codehaus.xfire.message.MessageWriter#writeValue(java.lang.Object)
     */
    public void writeValue(Object value, String ns, String attr)
    {
        try
        {
            if ( value != null )
                writer.writeAttribute( ns, attr, value.toString() );
        }
        catch ( XMLStreamException e )
        {
            // TODO: I'm still not sure what should happen here.
            throw new XFireRuntimeException("Error writing document.", e);
        }
    }
    
    /**
     * @see org.codehaus.xfire.message.MessageWriter#getWriter(java.lang.String)
     */
    public MessageWriter getChildWriter(String name)
    {
        return new LiteralWriter(writer, name, namespace);
    }

    public MessageWriter getChildWriter(String name, String ns)
    {
        return new LiteralWriter(writer, name, ns);
    }

    public MessageWriter getChildWriter(QName qname)
    {
        return new LiteralWriter(writer, 
                                 qname.getLocalPart(), 
                                 qname.getNamespaceURI(), 
                                 qname.getPrefix());
    }

    public String getNamespace()
    {
        return namespace;
    }

    public void setNamespace(String namespace)
    {
        this.namespace = namespace;
    }

    /**
     * @see org.codehaus.xfire.message.MessageWriter#writeValueAsCalendar(java.util.Calendar)
     */
    public void writeValueAsCalendar(Calendar calendar)
    {
         writeValue( DateUtils.formatDateTime( calendar.getTime() ) );
    }
    
    public void writeValueAsCalendar(Calendar calendar, String ns, String attr)
    {
         writeValue( DateUtils.formatDateTime( calendar.getTime() ) );
    }
    
    /**
     * @see org.codehaus.xfire.message.MessageWriter#writeValueAsInt(java.lang.Integer)
     */
    public void writeValueAsInt(Integer i)
    {
        writeValue( i.toString() );
    }

    public void writeValueAsInt(Integer i, String ns, String attr)
    {
        writeValue( i.toString() );
    }

    /**
     * @see org.codehaus.xfire.message.MessageWriter#writeValueAsDate(java.util.Date)
     */
    public void writeValueAsDateTime(Date date)
    {
         writeValue( DateUtils.formatDateTime(date) );
    }

    public void writeValueAsDateTime(Date date, String ns, String attr)
    {
         writeValue( ns, attr, DateUtils.formatDateTime(date) );
    }
    
    /**
     * @see org.codehaus.xfire.message.MessageWriter#writeValueAsDate(java.util.Date)
     */
    public void writeValueAsDate(Date date)
    {
        writeValue(DateUtils.formatDate( calendar.getTime() ));
    }

    public void writeValueAsDate(Date date, String ns, String attr)
    {
        writeValue(DateUtils.formatDate( calendar.getTime() ), ns, attr);
    }

	/**
	 * @see org.codehaus.xfire.message.MessageWriter#writeValueAsDouble(java.lang.Double)
	 */
	public void writeValueAsDouble(Double d)
	{
		writeValue( d.toString() );
	}

	public void writeValueAsDouble(Double d, String ns, String attr)
	{
		writeValue( d.toString(), ns, attr );
	}

	/**
	 * @see org.codehaus.xfire.message.MessageWriter#writeValueAsLong(java.lang.Long)
	 */
	public void writeValueAsLong(Long l)
	{
		writeValue( l.toString() );
	}

	public void writeValueAsLong(Long l, String ns, String attr)
	{
		writeValue( l.toString(), ns, attr );
	}

	/**
	 * @see org.codehaus.xfire.message.MessageWriter#writeValueAsFloat(java.lang.Float)
	 */
	public void writeValueAsFloat(Float f)
	{
		writeValue( f.toString() );
	}

	public void writeValueAsFloat(Float f, String ns, String attr)
	{
		writeValue( f.toString(), ns, attr );
	}

	/**
	 * @see org.codehaus.xfire.message.MessageWriter#writeValueAsBoolean(boolean)
	 */
	public void writeValueAsBoolean(boolean b)
	{
		writeValue( Boolean.toString(b) );
	}

	public void writeValueAsBoolean(boolean b, String ns, String attr)
	{
		writeValue( Boolean.toString(b), ns, attr );
	}

    public void close()
    {
        try
        {
            writer.writeEndElement();
        }
        catch ( XMLStreamException e )
        {
            // TODO: I'm still not sure what should happen here.
            throw new XFireRuntimeException("Error writing document.", e);
        }
    }

    public void flush() throws XMLStreamException
    {
        writer.flush();
    }

    public XMLStreamWriter getXMLStreamWriter()
    {
        return writer;
    }
}
