package org.codehaus.xfire.service.object;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.URL;

import javax.wsdl.Definition;
import javax.wsdl.factory.WSDLFactory;
import javax.wsdl.xml.WSDLReader;
import javax.xml.namespace.QName;

import org.codehaus.xfire.XFire;
import org.codehaus.xfire.fault.Soap11FaultHandler;
import org.codehaus.xfire.fault.Soap12FaultHandler;
import org.codehaus.xfire.handler.SoapHandler;
import org.codehaus.xfire.message.ObjectServiceHandler;
import org.codehaus.xfire.service.Service;
import org.codehaus.xfire.soap.Soap11;
import org.codehaus.xfire.soap.SoapConstants;
import org.codehaus.xfire.soap.SoapVersion;
import org.codehaus.xfire.type.DefaultTypeMappingRegistry;
import org.codehaus.xfire.type.TypeMapping;
import org.codehaus.xfire.type.TypeMappingRegistry;
import org.codehaus.xfire.util.NamespaceHelper;
import org.codehaus.xfire.wsdl.ResourceWSDL;
import org.codehaus.xfire.wsdl11.ObjectServiceVisitor;
import org.codehaus.xfire.wsdl11.builder.JavaWSDLBuilder;

/**
 * @author <a href="mailto:dan@envoisolutions.com">Dan Diephouse</a>
 */
public class ObjectServiceBuilder
    implements ServiceBuilder
{
    private TypeMappingRegistry registry;
    private XFire xfire;
    
    private Class invokerClass;
    private Class messageBridgeClass;
    
    public ObjectServiceBuilder()
    {
        registry = new DefaultTypeMappingRegistry(true);
    }
    
    public ObjectServiceBuilder(XFire xfire)
    {
        registry = new DefaultTypeMappingRegistry(true);
        
        this.xfire = xfire;
    }

    public ObjectServiceBuilder(XFire xfire, TypeMappingRegistry registry)
    {
        this.registry = registry;

        this.xfire = xfire;
    }

    /**
     * @param wsdlUrl
     * @return
     */
    public Service create(Class clazz, TypeMapping tm, URL wsdlUrl) throws Exception
    {
        WSDLReader reader = WSDLFactory.newInstance().newWSDLReader();
        Definition def = reader.readWSDL(wsdlUrl.toString());
        
        DefaultObjectService service = new DefaultObjectService();
        service.setServiceClass(clazz);
        service.setTypeMapping(tm);

        ObjectServiceVisitor builder = new ObjectServiceVisitor(def, service);
        builder.configure();

        service.setFaultHandler(new Soap11FaultHandler());
        service.setWSDL(new ResourceWSDL(wsdlUrl));
        
        SoapHandler handler = new SoapHandler(new ObjectServiceHandler());
        service.setServiceHandler(handler);
        
        xfire.getServiceRegistry().register(service);

        return service;
    }
    
    /**
     * @param clazz
     * @return
     */
    public Service create(Class clazz)
    {
        return create(clazz,
                      Soap11.getInstance(),
                      SoapConstants.STYLE_WRAPPED,
                      SoapConstants.USE_LITERAL);
    }

    /**
     * @param clazz
     * @param version
     * @param style
     * @param use
     * @return
     */
    public Service create(Class clazz, SoapVersion version, String style, String use)
    {
        String name = clazz.getName();
        int last = name.lastIndexOf(".");
        if ( last != -1 )
        {
            name = name.substring(last+1);
        }
        
        String ns = NamespaceHelper.makeNamespaceFromClassName(clazz.getName(), "http");
        return create(clazz, name, ns, version, style, use);
    }
    
    public Service create(Class clazz,
                          String name,
                          String namespace,
                          SoapVersion version,
                          String style,
                          String use)
    {
        DefaultObjectService service = new DefaultObjectService();

        service.setName(name);
        service.setDefaultNamespace(namespace);
        service.setServiceClass(clazz);
        
        service.setSoapVersion(version);
        service.setStyle(style);
        service.setUse(use);
        service.setAutoTyped(true);
        
        if (version instanceof Soap11)
        {
            service.setFaultHandler(new Soap11FaultHandler());
        }
        else
        {
            service.setFaultHandler(new Soap12FaultHandler());
        }
        
        if (xfire != null)
        {
            service.setWSDLBuilder(new JavaWSDLBuilder(xfire.getTransportManager()));
        }
        
        SoapHandler handler = new SoapHandler(new ObjectServiceHandler());
        service.setServiceHandler(handler);

        initializeTypeMapping(service);
        initializeOperations(service);
        
        if (xfire != null)
        {
            xfire.getServiceRegistry().register(service);
        }
        
        return service;
    }

    
    protected void initializeOperations(DefaultObjectService service)
    {
        final Method[] methods = service.getServiceClass().getDeclaredMethods();

        // TODO: go through superclasses, stopping at Object.class

        for (int i = 0; i < methods.length; i++)
        {
            final Method method = methods[i];
            final String methodName = method.getName();
            final int modifiers = method.getModifiers();

            if ( /*isAllowed(methodName) &&*/ Modifier.isPublic(modifiers)
                    && !Modifier.isStatic(modifiers))
            {
                addOperation(service, method);
            }
        }
    }

    protected void addOperation(DefaultObjectService service, final Method method)
    {
        final Operation op = new Operation(method);

        final Class[] paramClasses = method.getParameterTypes();

        final boolean isDoc = service.getStyle().equals(SoapConstants.STYLE_DOCUMENT);

        final TypeMapping mapping = service.getTypeMapping();
        
        for (int j = 0; j < paramClasses.length; j++)
        {
            String paramName = "";
            if (isDoc)
                paramName = method.getName();

            paramName = paramName + "in" + j;

            final QName q = new QName(service.getDefaultNamespace(), paramName);
            final Parameter p = new Parameter(q, paramClasses[j]);

            op.addInParameter(p);
        }

        final Class returnType = method.getReturnType();
        if (!returnType.isAssignableFrom(void.class))
        {
            String outName = "";
            if (isDoc)
                outName = method.getName();

            final QName q = new QName(service.getDefaultNamespace(), outName + "out");
            final Parameter outP = new Parameter(q, method.getReturnType());
            op.addOutParameter(outP);
        }

        service.addOperation(op);
    }

    /**
     * Creates a type mapping for this class and registers it with the
     * TypeMappingRegistry. This needs to be called before
     * initializeOperations().
     */
    protected void initializeTypeMapping(DefaultObjectService service)
    {
        String encodingStyle = service.getEncodingStyleURI();
            
        if (encodingStyle == null)
        {
            if (service.getUse().equals(SoapConstants.USE_ENCODED))
            {
                encodingStyle = service.getSoapVersion().getSoapEncodingStyle();
            }
            else
            {
                encodingStyle = SoapConstants.XSD;
            }
        }

        service.setEncodingStyleURI(encodingStyle);
        final TypeMapping tm = registry.createTypeMapping(encodingStyle, service.isAutoTyped());

        service.setTypeMapping(tm);
        registry.register(service.getDefaultNamespace(), tm);
    }

}
